"""
Managed index.

A managed Index - where the index is accessible via some API that
interfaces a managed service.

"""

import asyncio
import httpx
import os
import time
from typing import Any, Awaitable, Callable, List, Optional, Sequence, Type, Dict
from typing_extensions import override
from urllib.parse import quote_plus

from llama_cloud import (
    ManagedIngestionStatusResponse,
    PipelineCreate,
    PipelineCreateEmbeddingConfig,
    PipelineCreateTransformConfig,
    PipelineType,
    ProjectCreate,
    ManagedIngestionStatus,
    CloudDocumentCreate,
    CloudDocument,
    PipelineFileCreate,
    LlamaParseParameters,
)

from llama_index.core.base.base_query_engine import BaseQueryEngine
from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.constants import DEFAULT_APP_URL, DEFAULT_PROJECT_NAME
from llama_index.core.indices.managed.base import BaseManagedIndex

from llama_cloud.core.api_error import ApiError
from llama_index.core.ingestion.api_utils import (
    get_aclient,
    get_client,
)
from llama_index.core.schema import BaseNode, Document, TransformComponent
from llama_index.core.settings import Settings
from llama_index.core.storage.docstore.types import RefDocInfo
from .api_utils import (
    default_transform_config,
    resolve_project_and_pipeline,
)
import logging

logger = logging.getLogger(__name__)


class LlamaCloudIndex(BaseManagedIndex):
    """
    A managed index that stores documents in LlamaCloud.

    There are two main ways to use this index:

    1. Connect to an existing LlamaCloud index:
        ```python
        # Connect using index ID (same as pipeline ID)
        index = LlamaCloudIndex(id="<index_id>")

        # Or connect using index name
        index = LlamaCloudIndex(
            name="my_index",
            project_name="my_project",
            organization_id="my_org_id"
        )
        ```

    2. Create a new index with documents:
        ```python
        documents = [Document(...), Document(...)]
        index = LlamaCloudIndex.from_documents(
            documents,
            name="my_new_index",
            project_name="my_project",
            organization_id="my_org_id"
        )
        ```

    The index supports standard operations like retrieval and querying
    through the as_query_engine() and as_retriever() methods.
    """

    def __init__(
        self,
        # index identifier
        name: Optional[str] = None,
        pipeline_id: Optional[str] = None,
        index_id: Optional[str] = None,  # alias for pipeline_id
        id: Optional[str] = None,  # alias for pipeline_id
        # project identifier
        project_id: Optional[str] = None,
        project_name: str = DEFAULT_PROJECT_NAME,
        organization_id: Optional[str] = None,
        # connection params
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        app_url: Optional[str] = None,
        timeout: int = 60,
        httpx_client: Optional[httpx.Client] = None,
        async_httpx_client: Optional[httpx.AsyncClient] = None,
        # misc
        show_progress: bool = False,
        callback_manager: Optional[CallbackManager] = None,
        # deprecated
        nodes: Optional[List[BaseNode]] = None,
        transformations: Optional[List[TransformComponent]] = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Platform Index."""
        if sum([bool(id), bool(index_id), bool(pipeline_id), bool(name)]) != 1:
            raise ValueError(
                "Exactly one of `name`, `id`, `pipeline_id` or `index_id` must be provided to identify the index."
            )

        if nodes is not None:
            # TODO: How to handle uploading nodes without running transforms on them?
            raise ValueError("LlamaCloudIndex does not support nodes on initialization")

        if transformations is not None:
            raise ValueError(
                "Setting transformations is deprecated for LlamaCloudIndex, please use the `transform_config` and `embedding_config` parameters instead."
            )

        # initialize clients
        self._httpx_client = httpx_client
        self._async_httpx_client = async_httpx_client
        self._client = get_client(
            api_key=api_key,
            base_url=base_url,
            app_url=app_url,
            timeout=timeout,
            httpx_client=httpx_client,
        )
        self._aclient = get_aclient(
            api_key=api_key,
            base_url=base_url,
            app_url=app_url,
            timeout=timeout,
            httpx_client=async_httpx_client,
        )

        self.organization_id = organization_id
        pipeline_id = id or index_id or pipeline_id

        self.project, self.pipeline = resolve_project_and_pipeline(
            self._client, name, pipeline_id, project_name, project_id, organization_id
        )
        self.name = self.pipeline.name
        self.project_name = self.project.name

        self._api_key = api_key
        self._base_url = base_url
        self._app_url = app_url
        self._timeout = timeout
        self._show_progress = show_progress
        self._callback_manager = callback_manager or Settings.callback_manager

    @property
    def id(self) -> str:
        """Return the pipeline (aka index) ID."""
        return self.pipeline.id

    def _wait_for_resources(
        self,
        resource_ids: Sequence[str],
        get_status_fn: Callable[[str], ManagedIngestionStatusResponse],
        resource_name: str,
        verbose: bool,
        raise_on_error: bool,
        sleep_interval: float,
    ) -> None:
        """
        Poll `get_status_fn` until every id in `resource_ids` is finished.

        Args:
            resource_ids: Iterable of resource ids to watch.
            get_status_fn: Callable that maps a resource id → ManagedIngestionStatus.
            resource_name: Text used in log / error messages: "file", "document", ….
            verbose: Print a progress bar.
            raise_on_error: Whether to raise on ManagedIngestionStatus.ERROR.
            sleep_interval: Seconds between polls (min 0.5 s to avoid rate-limits).

        """
        if not resource_ids:  # nothing to do
            return

        if verbose:
            print(
                f"Loading {resource_name}{'s' if len(resource_ids) > 1 else ''}",
            )

        pending: set[str] = set(resource_ids)
        while pending:
            finished: set[str] = set()
            for rid in pending:
                try:
                    status_response = get_status_fn(rid)
                    status = status_response.status
                    if status in (
                        ManagedIngestionStatus.NOT_STARTED,
                        ManagedIngestionStatus.IN_PROGRESS,
                    ):
                        continue  # still working

                    if status == ManagedIngestionStatus.ERROR:
                        if verbose:
                            print(
                                f"{resource_name.capitalize()} ingestion failed for {rid}"
                            )
                        if raise_on_error:
                            raise ValueError(
                                f"{resource_name.capitalize()} ingestion failed for {rid}"
                            )

                    finished.add(rid)
                    if verbose:
                        print(
                            f"{resource_name.capitalize()} ingestion finished for {rid}"
                        )

                except httpx.HTTPStatusError as e:
                    if e.response.status_code in (429, 500, 502, 503, 504):
                        pass
                    else:
                        raise

            pending -= finished

            if pending:
                time.sleep(sleep_interval)

        if verbose:
            print("Done!")

    async def _await_for_resources(
        self,
        resource_ids: Sequence[str],
        get_status_fn: Callable[[str], Awaitable[ManagedIngestionStatusResponse]],
        resource_name: str,
        verbose: bool,
        raise_on_error: bool,
        sleep_interval: float,
    ) -> None:
        """
        Poll `get_status_fn` until every id in `resource_ids` is finished.

        Args:
            resource_ids: Iterable of resource ids to watch.
            get_status_fn: Callable that maps a resource id → ManagedIngestionStatus.
            resource_name: Text used in log / error messages: "file", "document", ….
            verbose: Print a progress bar.
            raise_on_error: Whether to raise on ManagedIngestionStatus.ERROR.
            sleep_interval: Seconds between polls (min 0.5 s to avoid rate-limits).

        """
        if not resource_ids:  # nothing to do
            return

        if verbose:
            print(
                f"Loading {resource_name}{'s' if len(resource_ids) > 1 else ''}",
            )

        pending: set[str] = set(resource_ids)
        while pending:
            finished: set[str] = set()
            for rid in pending:
                try:
                    status_response = await get_status_fn(rid)
                    status = status_response.status
                    if status in (
                        ManagedIngestionStatus.NOT_STARTED,
                        ManagedIngestionStatus.IN_PROGRESS,
                    ):
                        continue  # still working

                    if status == ManagedIngestionStatus.ERROR:
                        if verbose:
                            print(
                                f"{resource_name.capitalize()} ingestion failed for {rid}"
                            )
                        if raise_on_error:
                            raise ValueError(
                                f"{resource_name.capitalize()} ingestion failed for {rid}"
                            )

                    finished.add(rid)
                    if verbose:
                        print(
                            f"{resource_name.capitalize()} ingestion finished for {rid}"
                        )

                except httpx.HTTPStatusError as e:
                    if e.response.status_code in (429, 500, 502, 503, 504):
                        pass
                    else:
                        raise

            pending -= finished

            if pending:
                await asyncio.sleep(sleep_interval)

        if verbose:
            print("Done!")

    def wait_for_completion(
        self,
        file_ids: Optional[Sequence[str]] = None,
        doc_ids: Optional[Sequence[str]] = None,
        verbose: bool = False,
        raise_on_partial_success: bool = False,
        raise_on_error: bool = False,
        sleep_interval: float = 1.0,
    ) -> Optional[ManagedIngestionStatusResponse]:
        """
        Block until the requested ingestion work is finished.

        - If `file_ids` is given → wait for those files.
        - If `doc_ids` is given → wait for those documents.
        - If neither is given → wait for the pipeline itself last so that retrieval works.
        - Always waits for the pipeline itself last so that retrieval works.

        Returns the final PipelineStatus response (or None if only waiting on
        files / documents).
        """
        # Batch of files (if any)
        if file_ids:
            self._wait_for_resources(
                file_ids,
                lambda fid: self._client.pipelines.get_pipeline_file_status(
                    pipeline_id=self.pipeline.id, file_id=fid
                ),
                resource_name="file",
                verbose=verbose,
                raise_on_error=raise_on_error,
                sleep_interval=sleep_interval,
            )

        # Batch of documents (if any)
        if doc_ids:
            self._wait_for_resources(
                doc_ids,
                lambda did: self._client.pipelines.get_pipeline_document_status(
                    pipeline_id=self.pipeline.id,
                    document_id=quote_plus(quote_plus(did)),
                ),
                resource_name="document",
                verbose=verbose,
                raise_on_error=raise_on_error,
                sleep_interval=sleep_interval,
            )

        # Finally, wait for the pipeline
        if verbose:
            print(f"Syncing pipeline {self.pipeline.id}")

        status_response: Optional[ManagedIngestionStatusResponse] = None
        while True:
            try:
                status_response = self._client.pipelines.get_pipeline_status(
                    pipeline_id=self.pipeline.id
                )
                status = status_response.status
            except httpx.HTTPStatusError as e:
                if e.response.status_code in (429, 500, 502, 503, 504):
                    time.sleep(sleep_interval)
                    continue
                else:
                    raise

            if status == ManagedIngestionStatus.ERROR or (
                raise_on_partial_success
                and status == ManagedIngestionStatus.PARTIAL_SUCCESS
            ):
                raise ValueError(
                    f"Pipeline ingestion failed for {self.pipeline.id}. "
                    f"Details: {status_response.json()}"
                )

            if status in (
                ManagedIngestionStatus.NOT_STARTED,
                ManagedIngestionStatus.IN_PROGRESS,
            ):
                if verbose:
                    print(".", end="")
                time.sleep(sleep_interval)
            else:
                if verbose:
                    print("Done!")

                return status_response

    async def await_for_completion(
        self,
        file_ids: Optional[Sequence[str]] = None,
        doc_ids: Optional[Sequence[str]] = None,
        verbose: bool = False,
        raise_on_partial_success: bool = False,
        raise_on_error: bool = False,
        sleep_interval: float = 1.0,
    ) -> Optional[ManagedIngestionStatusResponse]:
        """
        Block until the requested ingestion work is finished.

        - If `file_ids` is given → wait for those files.
        - If `doc_ids` is given → wait for those documents.
        - If neither is given → wait for the pipeline itself last so that retrieval works.
        - Always waits for the pipeline itself last so that retrieval works.

        Returns the final PipelineStatus response (or None if only waiting on
        files / documents).
        """
        # Batch of files (if any)
        if file_ids:
            await self._await_for_resources(
                file_ids,
                lambda fid: self._aclient.pipelines.get_pipeline_file_status(
                    pipeline_id=self.pipeline.id, file_id=fid
                ),
                resource_name="file",
                verbose=verbose,
                raise_on_error=raise_on_error,
                sleep_interval=sleep_interval,
            )

        # Batch of documents (if any)
        if doc_ids:
            await self._await_for_resources(
                doc_ids,
                lambda did: self._aclient.pipelines.get_pipeline_document_status(
                    pipeline_id=self.pipeline.id,
                    document_id=quote_plus(quote_plus(did)),
                ),
                resource_name="document",
                verbose=verbose,
                raise_on_error=raise_on_error,
                sleep_interval=sleep_interval,
            )

        # Finally, wait for the pipeline
        if verbose:
            print(f"Syncing pipeline {self.pipeline.id}")

        status_response: Optional[ManagedIngestionStatusResponse] = None
        while True:
            try:
                status_response = await self._aclient.pipelines.get_pipeline_status(
                    pipeline_id=self.pipeline.id
                )
                status = status_response.status
            except httpx.HTTPStatusError as e:
                if e.response.status_code in (429, 500, 502, 503, 504):
                    await asyncio.sleep(sleep_interval)
                    continue
                else:
                    raise

            if status == ManagedIngestionStatus.ERROR or (
                raise_on_partial_success
                and status == ManagedIngestionStatus.PARTIAL_SUCCESS
            ):
                raise ValueError(
                    f"Pipeline ingestion failed for {self.pipeline.id}. "
                    f"Details: {status_response.json()}"
                )

            if status in (
                ManagedIngestionStatus.NOT_STARTED,
                ManagedIngestionStatus.IN_PROGRESS,
            ):
                if verbose:
                    print(".", end="")
                await asyncio.sleep(sleep_interval)
            else:
                if verbose:
                    print("Done!")

                return status_response

    @classmethod
    def create_index(
        cls: Type["LlamaCloudIndex"],
        name: str,
        project_name: str = DEFAULT_PROJECT_NAME,
        organization_id: Optional[str] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        app_url: Optional[str] = None,
        timeout: int = 60,
        verbose: bool = False,
        # ingestion configs
        embedding_config: Optional[PipelineCreateEmbeddingConfig] = None,
        transform_config: Optional[PipelineCreateTransformConfig] = None,
        llama_parse_parameters: Optional[LlamaParseParameters] = None,
        **kwargs: Any,
    ) -> "LlamaCloudIndex":
        """Create a new LlamaCloud managed index."""
        app_url = app_url or os.environ.get("LLAMA_CLOUD_APP_URL", DEFAULT_APP_URL)
        client = get_client(api_key, base_url, app_url, timeout)

        # create project if it doesn't exist
        project = client.projects.upsert_project(
            organization_id=organization_id, request=ProjectCreate(name=project_name)
        )
        if project.id is None:
            raise ValueError(f"Failed to create/get project {project_name}")

        if verbose:
            print(f"Created project {project.id} with name {project.name}")

        # create pipeline
        pipeline_create = PipelineCreate(
            name=name,
            pipeline_type=PipelineType.MANAGED,
            embedding_config=embedding_config,  # If it's None, the default embedding config will be used
            transform_config=transform_config or default_transform_config(),
            llama_parse_parameters=llama_parse_parameters or LlamaParseParameters(),
        )
        pipeline = client.pipelines.upsert_pipeline(
            project_id=project.id, request=pipeline_create
        )
        if pipeline.id is None:
            raise ValueError(f"Failed to create/get pipeline {name}")
        if verbose:
            print(f"Created pipeline {pipeline.id} with name {pipeline.name}")

        return cls(
            name,
            project_name=project.name,
            organization_id=project.organization_id,
            api_key=api_key,
            base_url=base_url,
            app_url=app_url,
            timeout=timeout,
            **kwargs,
        )

    @classmethod
    async def acreate_index(
        cls: Type["LlamaCloudIndex"],
        name: str,
        project_name: str = DEFAULT_PROJECT_NAME,
        organization_id: Optional[str] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        app_url: Optional[str] = None,
        timeout: int = 60,
        verbose: bool = False,
        # ingestion configs
        embedding_config: Optional[PipelineCreateEmbeddingConfig] = None,
        transform_config: Optional[PipelineCreateTransformConfig] = None,
        llama_parse_parameters: Optional[LlamaParseParameters] = None,
        **kwargs: Any,
    ) -> "LlamaCloudIndex":
        """Create a new LlamaCloud managed index."""
        app_url = app_url or os.environ.get("LLAMA_CLOUD_APP_URL", DEFAULT_APP_URL)
        aclient = get_aclient(api_key, base_url, app_url, timeout)

        # create project if it doesn't exist
        project = await aclient.projects.upsert_project(
            organization_id=organization_id, request=ProjectCreate(name=project_name)
        )
        if project.id is None:
            raise ValueError(f"Failed to create/get project {project_name}")

        if verbose:
            print(f"Created project {project.id} with name {project.name}")

        # create pipeline
        pipeline_create = PipelineCreate(
            name=name,
            pipeline_type=PipelineType.MANAGED,
            embedding_config=embedding_config,  # If it's None, the default embedding config will be used
            transform_config=transform_config or default_transform_config(),
            llama_parse_parameters=llama_parse_parameters or LlamaParseParameters(),
        )
        pipeline = await aclient.pipelines.upsert_pipeline(
            project_id=project.id, request=pipeline_create
        )
        if pipeline.id is None:
            raise ValueError(f"Failed to create/get pipeline {name}")
        if verbose:
            print(f"Created pipeline {pipeline.id} with name {pipeline.name}")

        return cls(
            name,
            project_name=project.name,
            organization_id=project.organization_id,
            api_key=api_key,
            base_url=base_url,
            app_url=app_url,
            timeout=timeout,
            **kwargs,
        )

    @classmethod
    def from_documents(  # type: ignore
        cls: Type["LlamaCloudIndex"],
        documents: List[Document],
        name: str,
        project_name: str = DEFAULT_PROJECT_NAME,
        organization_id: Optional[str] = None,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        app_url: Optional[str] = None,
        timeout: int = 60,
        verbose: bool = False,
        raise_on_error: bool = False,
        # ingestion configs
        embedding_config: Optional[PipelineCreateEmbeddingConfig] = None,
        transform_config: Optional[PipelineCreateTransformConfig] = None,
        # deprecated
        transformations: Optional[List[TransformComponent]] = None,
        **kwargs: Any,
    ) -> "LlamaCloudIndex":
        """Build a LlamaCloud managed index from a sequence of documents."""
        index = cls.create_index(
            name=name,
            project_name=project_name,
            organization_id=organization_id,
            api_key=api_key,
            base_url=base_url,
            app_url=app_url,
            timeout=timeout,
            verbose=verbose,
            embedding_config=embedding_config,
            transform_config=transform_config,
        )

        app_url = app_url or os.environ.get("LLAMA_CLOUD_APP_URL", DEFAULT_APP_URL)
        client = get_client(api_key, base_url, app_url, timeout)

        # this kicks off document ingestion
        upserted_documents = client.pipelines.upsert_batch_pipeline_documents(
            pipeline_id=index.pipeline.id,
            request=[
                CloudDocumentCreate(
                    text=doc.text,
                    metadata=doc.metadata,
                    excluded_embed_metadata_keys=doc.excluded_embed_metadata_keys,
                    excluded_llm_metadata_keys=doc.excluded_llm_metadata_keys,
                    id=doc.id_,
                )
                for doc in documents
            ],
        )

        doc_ids = [doc.id for doc in upserted_documents]
        index.wait_for_completion(
            doc_ids=doc_ids, verbose=verbose, raise_on_error=raise_on_error
        )

        print(
            f"Find your index at {app_url}/project/{index.project.id}/deploy/{index.pipeline.id}"
        )

        return index

    def as_retriever(self, **kwargs: Any) -> BaseRetriever:
        """Return a Retriever for this managed index."""
        from .retriever import (
            LlamaCloudRetriever,
        )

        similarity_top_k = kwargs.pop("similarity_top_k", None)
        dense_similarity_top_k = kwargs.pop("dense_similarity_top_k", None)
        if similarity_top_k is not None:
            dense_similarity_top_k = similarity_top_k

        return LlamaCloudRetriever(
            project_id=self.project.id,
            pipeline_id=self.pipeline.id,
            api_key=self._api_key,
            base_url=self._base_url,
            app_url=self._app_url,
            timeout=self._timeout,
            organization_id=self.organization_id,
            dense_similarity_top_k=dense_similarity_top_k,
            httpx_client=self._httpx_client,
            async_httpx_client=self._async_httpx_client,
            **kwargs,
        )

    @override
    def as_query_engine(self, **kwargs: Any) -> BaseQueryEngine:
        from llama_index.core.query_engine.retriever_query_engine import (
            RetrieverQueryEngine,
        )

        kwargs["retriever"] = self.as_retriever(**kwargs)
        return RetrieverQueryEngine.from_args(**kwargs)

    @property
    def ref_doc_info(self, batch_size: int = 100) -> Dict[str, RefDocInfo]:
        """Retrieve a dict mapping of ingested documents and their metadata. The nodes list is empty."""
        pipeline_id = self.pipeline.id
        pipeline_documents: List[CloudDocument] = []
        skip = 0
        limit = batch_size
        while True:
            batch = self._client.pipelines.list_pipeline_documents(
                pipeline_id=pipeline_id,
                skip=skip,
                limit=limit,
            )
            if not batch:
                break
            pipeline_documents.extend(batch)
            skip += limit
        return {
            doc.id: RefDocInfo(metadata=doc.metadata, node_ids=[])
            for doc in pipeline_documents
        }

    def insert(
        self, document: Document, verbose: bool = False, **insert_kwargs: Any
    ) -> None:
        """Insert a document."""
        with self._callback_manager.as_trace("insert"):
            upserted_documents = self._client.pipelines.create_batch_pipeline_documents(
                pipeline_id=self.pipeline.id,
                request=[
                    CloudDocumentCreate(
                        text=document.text,
                        metadata=document.metadata,
                        excluded_embed_metadata_keys=document.excluded_embed_metadata_keys,
                        excluded_llm_metadata_keys=document.excluded_llm_metadata_keys,
                        id=document.id_,
                    )
                ],
            )
            upserted_document = upserted_documents[0]
            self.wait_for_completion(
                doc_ids=[upserted_document.id], verbose=verbose, raise_on_error=True
            )

    async def ainsert(
        self, document: Document, verbose: bool = False, **insert_kwargs: Any
    ) -> None:
        """Insert a document."""
        with self._callback_manager.as_trace("insert"):
            upserted_documents = await self._aclient.pipelines.create_batch_pipeline_documents(
                pipeline_id=self.pipeline.id,
                request=[
                    CloudDocumentCreate(
                        text=document.text,
                        metadata=document.metadata,
                        excluded_embed_metadata_keys=document.excluded_embed_metadata_keys,
                        excluded_llm_metadata_keys=document.excluded_llm_metadata_keys,
                        id=document.id_,
                    )
                ],
            )
            upserted_document = upserted_documents[0]
            await self.await_for_completion(
                doc_ids=[upserted_document.id], verbose=verbose, raise_on_error=True
            )

    def update_ref_doc(
        self, document: Document, verbose: bool = False, **update_kwargs: Any
    ) -> None:
        """Upserts a document and its corresponding nodes."""
        with self._callback_manager.as_trace("update"):
            upserted_documents = self._client.pipelines.upsert_batch_pipeline_documents(
                pipeline_id=self.pipeline.id,
                request=[
                    CloudDocumentCreate(
                        text=document.text,
                        metadata=document.metadata,
                        excluded_embed_metadata_keys=document.excluded_embed_metadata_keys,
                        excluded_llm_metadata_keys=document.excluded_llm_metadata_keys,
                        id=document.id_,
                    )
                ],
            )
            upserted_document = upserted_documents[0]
            self.wait_for_completion(
                doc_ids=[upserted_document.id], verbose=verbose, raise_on_error=True
            )

    async def aupdate_ref_doc(
        self, document: Document, verbose: bool = False, **update_kwargs: Any
    ) -> None:
        """Upserts a document and its corresponding nodes."""
        with self._callback_manager.as_trace("update"):
            upserted_documents = await self._aclient.pipelines.upsert_batch_pipeline_documents(
                pipeline_id=self.pipeline.id,
                request=[
                    CloudDocumentCreate(
                        text=document.text,
                        metadata=document.metadata,
                        excluded_embed_metadata_keys=document.excluded_embed_metadata_keys,
                        excluded_llm_metadata_keys=document.excluded_llm_metadata_keys,
                        id=document.id_,
                    )
                ],
            )
            upserted_document = upserted_documents[0]
            await self.await_for_completion(
                doc_ids=[upserted_document.id], verbose=verbose, raise_on_error=True
            )

    def refresh_ref_docs(
        self, documents: Sequence[Document], **update_kwargs: Any
    ) -> List[bool]:
        """Refresh an index with documents that have changed."""
        with self._callback_manager.as_trace("refresh"):
            upserted_documents = self._client.pipelines.upsert_batch_pipeline_documents(
                pipeline_id=self.pipeline.id,
                request=[
                    CloudDocumentCreate(
                        text=doc.text,
                        metadata=doc.metadata,
                        excluded_embed_metadata_keys=doc.excluded_embed_metadata_keys,
                        excluded_llm_metadata_keys=doc.excluded_llm_metadata_keys,
                        id=doc.id_,
                    )
                    for doc in documents
                ],
            )
            doc_ids = [doc.id for doc in upserted_documents]
            self.wait_for_completion(doc_ids=doc_ids, verbose=True, raise_on_error=True)
            return [True] * len(doc_ids)

    async def arefresh_ref_docs(
        self, documents: Sequence[Document], **update_kwargs: Any
    ) -> List[bool]:
        """Refresh an index with documents that have changed."""
        with self._callback_manager.as_trace("refresh"):
            upserted_documents = await self._aclient.pipelines.upsert_batch_pipeline_documents(
                pipeline_id=self.pipeline.id,
                request=[
                    CloudDocumentCreate(
                        text=doc.text,
                        metadata=doc.metadata,
                        excluded_embed_metadata_keys=doc.excluded_embed_metadata_keys,
                        excluded_llm_metadata_keys=doc.excluded_llm_metadata_keys,
                        id=doc.id_,
                    )
                    for doc in documents
                ],
            )
            doc_ids = [doc.id for doc in upserted_documents]
            await self.await_for_completion(
                doc_ids=doc_ids, verbose=True, raise_on_error=True
            )
            return [True] * len(doc_ids)

    def delete_ref_doc(
        self,
        ref_doc_id: str,
        delete_from_docstore: bool = False,
        verbose: bool = False,
        raise_if_not_found: bool = False,
        **delete_kwargs: Any,
    ) -> None:
        """Delete a document and its nodes by using ref_doc_id."""
        try:
            # we have to quote the ref_doc_id twice because it is used as a path parameter
            self._client.pipelines.delete_pipeline_document(
                pipeline_id=self.pipeline.id,
                document_id=quote_plus(quote_plus(ref_doc_id)),
            )
        except ApiError as e:
            if e.status_code == 404 and not raise_if_not_found:
                logger.warning(f"ref_doc_id {ref_doc_id} not found, nothing deleted.")
            else:
                raise

        # we have to wait for the pipeline instead of the document, because the document is already deleted
        self.wait_for_completion(verbose=verbose, raise_on_partial_success=False)

    async def adelete_ref_doc(
        self,
        ref_doc_id: str,
        delete_from_docstore: bool = False,
        verbose: bool = False,
        raise_if_not_found: bool = False,
        **delete_kwargs: Any,
    ) -> None:
        """Delete a document and its nodes by using ref_doc_id."""
        try:
            # we have to quote the ref_doc_id twice because it is used as a path parameter
            await self._aclient.pipelines.delete_pipeline_document(
                pipeline_id=self.pipeline.id,
                document_id=quote_plus(quote_plus(ref_doc_id)),
            )
        except ApiError as e:
            if e.status_code == 404 and not raise_if_not_found:
                logger.warning(f"ref_doc_id {ref_doc_id} not found, nothing deleted.")
            else:
                raise

        # we have to wait for the pipeline instead of the document, because the document is already deleted
        await self.await_for_completion(verbose=verbose, raise_on_partial_success=False)

    def upload_file(
        self,
        file_path: str,
        verbose: bool = False,
        wait_for_ingestion: bool = True,
        raise_on_error: bool = False,
    ) -> str:
        """Upload a file to the index."""
        with open(file_path, "rb") as f:
            file = self._client.files.upload_file(
                project_id=self.project.id, upload_file=f
            )
            if verbose:
                print(f"Uploaded file {file.id} with name {file.name}")

        # Add file to pipeline
        pipeline_file_create = PipelineFileCreate(file_id=file.id)
        self._client.pipelines.add_files_to_pipeline_api(
            pipeline_id=self.pipeline.id, request=[pipeline_file_create]
        )

        if wait_for_ingestion:
            self.wait_for_completion(
                file_ids=[file.id], verbose=verbose, raise_on_error=raise_on_error
            )
        return file.id

    async def aupload_file(
        self,
        file_path: str,
        verbose: bool = False,
        wait_for_ingestion: bool = True,
        raise_on_error: bool = False,
    ) -> str:
        """Upload a file to the index."""
        with open(file_path, "rb") as f:
            file = await self._aclient.files.upload_file(
                project_id=self.project.id, upload_file=f
            )
            if verbose:
                print(f"Uploaded file {file.id} with name {file.name}")

        # Add file to pipeline
        pipeline_file_create = PipelineFileCreate(file_id=file.id)
        await self._aclient.pipelines.add_files_to_pipeline_api(
            pipeline_id=self.pipeline.id, request=[pipeline_file_create]
        )

        if wait_for_ingestion:
            await self.await_for_completion(
                file_ids=[file.id], verbose=verbose, raise_on_error=raise_on_error
            )

        return file.id

    def upload_file_from_url(
        self,
        file_name: str,
        url: str,
        proxy_url: Optional[str] = None,
        request_headers: Optional[Dict[str, str]] = None,
        verify_ssl: bool = True,
        follow_redirects: bool = True,
        verbose: bool = False,
        wait_for_ingestion: bool = True,
        raise_on_error: bool = False,
    ) -> str:
        """Upload a file from a URL to the index."""
        file = self._client.files.upload_file_from_url(
            project_id=self.project.id,
            name=file_name,
            url=url,
            proxy_url=proxy_url,
            request_headers=request_headers,
            verify_ssl=verify_ssl,
            follow_redirects=follow_redirects,
        )
        if verbose:
            print(f"Uploaded file {file.id} with ID {file.id}")

        # Add file to pipeline
        pipeline_file_create = PipelineFileCreate(file_id=file.id)
        self._client.pipelines.add_files_to_pipeline_api(
            pipeline_id=self.pipeline.id, request=[pipeline_file_create]
        )

        if wait_for_ingestion:
            self.wait_for_completion(
                file_ids=[file.id], verbose=verbose, raise_on_error=raise_on_error
            )
        return file.id

    async def aupload_file_from_url(
        self,
        file_name: str,
        url: str,
        proxy_url: Optional[str] = None,
        request_headers: Optional[Dict[str, str]] = None,
        verify_ssl: bool = True,
        follow_redirects: bool = True,
        verbose: bool = False,
        wait_for_ingestion: bool = True,
        raise_on_error: bool = False,
    ) -> str:
        """Upload a file from a URL to the index."""
        file = await self._aclient.files.upload_file_from_url(
            project_id=self.project.id,
            name=file_name,
            url=url,
            proxy_url=proxy_url,
            request_headers=request_headers,
            verify_ssl=verify_ssl,
            follow_redirects=follow_redirects,
        )
        if verbose:
            print(f"Uploaded file {file.id} with ID {file.id}")

        # Add file to pipeline
        pipeline_file_create = PipelineFileCreate(file_id=file.id)
        await self._aclient.pipelines.add_files_to_pipeline_api(
            pipeline_id=self.pipeline.id, request=[pipeline_file_create]
        )

        if wait_for_ingestion:
            await self.await_for_completion(
                file_ids=[file.id], verbose=verbose, raise_on_error=raise_on_error
            )

        return file.id

    # Nodes related methods (not implemented for LlamaCloudIndex)

    def _insert(self, nodes: Sequence[BaseNode], **insert_kwargs: Any) -> None:
        """Index-specific logic for inserting nodes to the index struct."""
        raise NotImplementedError("_insert not implemented for LlamaCloudIndex.")

    @override
    def build_index_from_nodes(self, nodes: Sequence[BaseNode]) -> None:
        """Build the index from nodes."""
        raise NotImplementedError(
            "build_index_from_nodes not implemented for LlamaCloudIndex."
        )

    def insert_nodes(self, nodes: Sequence[BaseNode], **insert_kwargs: Any) -> None:
        """Insert a set of nodes."""
        raise NotImplementedError("insert_nodes not implemented for LlamaCloudIndex.")

    def delete_nodes(
        self,
        node_ids: List[str],
        delete_from_docstore: bool = False,
        **delete_kwargs: Any,
    ) -> None:
        """Delete a set of nodes."""
        raise NotImplementedError("delete_nodes not implemented for LlamaCloudIndex.")



import httpx
import itertools
import logging
import os
from enum import Enum
from pathlib import Path
from datetime import datetime
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    retry_if_exception,
    before_sleep_log,
)
from typing import Any, Iterable, Iterator, Optional, List, cast

logger = logging.getLogger(__name__)

# Asyncio error messages
nest_asyncio_err = "cannot be called from a running event loop"
nest_asyncio_msg = "The event loop is already running. Add `import nest_asyncio; nest_asyncio.apply()` to your code to fix this issue."


class ResultType(str, Enum):
    """The result type for the parser."""

    TXT = "text"
    MD = "markdown"
    JSON = "json"
    STRUCTURED = "structured"


class ParsingMode(str, Enum):
    """The parsing mode for the parser."""

    parse_page_without_llm = "parse_page_without_llm"
    parse_page_with_llm = "parse_page_with_llm"
    parse_page_with_lvm = "parse_page_with_lvm"
    parse_page_with_agent = "parse_page_with_agent"
    parse_document_with_llm = "parse_document_with_llm"
    parse_document_with_agent = "parse_document_with_agent"


class FailedPageMode(str, Enum):
    """
    Enum for representing the different available page error handling modes
    """

    raw_text = "raw_text"
    blank_page = "blank_page"
    error_message = "error_message"


class Language(str, Enum):
    BAZA = "abq"
    ADYGHE = "ady"
    AFRIKAANS = "af"
    ANGIKA = "ang"
    ARABIC = "ar"
    ASSAMESE = "as"
    AVAR = "ava"
    AZERBAIJANI = "az"
    BELARUSIAN = "be"
    BULGARIAN = "bg"
    BIHARI = "bh"
    BHOJPURI = "bho"
    BENGALI = "bn"
    BOSNIAN = "bs"
    SIMPLIFIED_CHINESE = "ch_sim"
    TRADITIONAL_CHINESE = "ch_tra"
    CHECHEN = "che"
    CZECH = "cs"
    WELSH = "cy"
    DANISH = "da"
    DARGWA = "dar"
    GERMAN = "de"
    ENGLISH = "en"
    SPANISH = "es"
    ESTONIAN = "et"
    PERSIAN_FARSI = "fa"
    FRENCH = "fr"
    IRISH = "ga"
    GOAN_KONKANI = "gom"
    HINDI = "hi"
    CROATIAN = "hr"
    HUNGARIAN = "hu"
    INDONESIAN = "id"
    INGUSH = "inh"
    ICELANDIC = "is"
    ITALIAN = "it"
    JAPANESE = "ja"
    KABARDIAN = "kbd"
    KANNADA = "kn"
    KOREAN = "ko"
    KURDISH = "ku"
    LATIN = "la"
    LAK = "lbe"
    LEZGHIAN = "lez"
    LITHUANIAN = "lt"
    LATVIAN = "lv"
    MAGAHI = "mah"
    MAITHILI = "mai"
    MAORI = "mi"
    MONGOLIAN = "mn"
    MARATHI = "mr"
    MALAY = "ms"
    MALTESE = "mt"
    NEPALI = "ne"
    NEWARI = "new"
    DUTCH = "nl"
    NORWEGIAN = "no"
    OCCITAN = "oc"
    PALI = "pi"
    POLISH = "pl"
    PORTUGUESE = "pt"
    ROMANIAN = "ro"
    RUSSIAN = "ru"
    SERBIAN_CYRILLIC = "rs_cyrillic"
    SERBIAN_LATIN = "rs_latin"
    NAGPURI = "sck"
    SLOVAK = "sk"
    SLOVENIAN = "sl"
    ALBANIAN = "sq"
    SWEDISH = "sv"
    SWAHILI = "sw"
    TAMIL = "ta"
    TABASSARAN = "tab"
    TELUGU = "te"
    THAI = "th"
    TAJIK = "tjk"
    TAGALOG = "tl"
    TURKISH = "tr"
    UYGHUR = "ug"
    UKRAINIAN = "uk"
    URDU = "ur"
    UZBEK = "uz"
    VIETNAMESE = "vi"


SUPPORTED_FILE_TYPES = [
    ".pdf",
    # document and presentations
    ".602",
    ".abw",
    ".cgm",
    ".cwk",
    ".doc",
    ".docx",
    ".docm",
    ".dot",
    ".dotm",
    ".hwp",
    ".key",
    ".lwp",
    ".mw",
    ".mcw",
    ".pages",
    ".pbd",
    ".ppt",
    ".pptm",
    ".pptx",
    ".pot",
    ".potm",
    ".potx",
    ".rtf",
    ".sda",
    ".sdd",
    ".sdp",
    ".sdw",
    ".sgl",
    ".sti",
    ".sxi",
    ".sxw",
    ".stw",
    ".sxg",
    ".txt",
    ".uof",
    ".uop",
    ".uot",
    ".vor",
    ".wpd",
    ".wps",
    ".xml",
    ".zabw",
    ".epub",
    # images
    ".jpg",
    ".jpeg",
    ".png",
    ".gif",
    ".bmp",
    ".svg",
    ".tiff",
    ".webp",
    # web
    ".htm",
    ".html",
    # spreadsheets
    ".xlsx",
    ".xls",
    ".xlsm",
    ".xlsb",
    ".xlw",
    ".csv",
    ".dif",
    ".sylk",
    ".slk",
    ".prn",
    ".numbers",
    ".et",
    ".ods",
    ".fods",
    ".uos1",
    ".uos2",
    ".dbf",
    ".wk1",
    ".wk2",
    ".wk3",
    ".wk4",
    ".wks",
    ".123",
    ".wq1",
    ".wq2",
    ".wb1",
    ".wb2",
    ".wb3",
    ".qpw",
    ".xlr",
    ".eth",
    ".tsv",
    ".mp3",
    ".mp4",
    ".mpeg",
    ".mpga",
    ".m4a",
    ".wav",
    ".webm",
]


def should_retry(exception: Exception) -> bool:
    """Check if the exception should be retried.

    Args:
        exception: The exception to check.
    """
    # Retry on connection errors (network issues)
    if isinstance(
        exception,
        (
            httpx.ConnectError,
            httpx.ConnectTimeout,
            httpx.ReadTimeout,
            httpx.WriteTimeout,
            httpx.RemoteProtocolError,
        ),
    ):
        return True

    # Retry on specific HTTP status codes
    if isinstance(exception, httpx.HTTPStatusError):
        status_code = exception.response.status_code
        # Retry on rate limiting or temporary server errors
        return status_code in (429, 500, 502, 503, 504)

    return False


async def make_api_request(
    client: httpx.AsyncClient,
    method: str,
    url: str,
    timeout: float = 60.0,
    max_retries: int = 5,
    **httpx_kwargs: Any,
) -> httpx.Response:
    """Make an retrying API request to the LlamaParse API.

    Args:
        client: The httpx.AsyncClient to use for the request.
        url: The URL to request.
        headers: The headers to include in the request.
        timeout: The timeout for the request.
        max_retries: The maximum number of retries for the request.
    """

    @retry(
        stop=stop_after_attempt(max_retries),
        wait=wait_exponential(multiplier=1, min=4, max=timeout),
        retry=retry_if_exception(should_retry),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )
    async def _make_request(url: str, **httpx_kwargs: Any) -> httpx.Response:
        if method == "GET":
            response = await client.get(url, **httpx_kwargs)
        elif method == "POST":
            response = await client.post(url, **httpx_kwargs)
        else:
            raise ValueError(f"Invalid method: {method}")
        response.raise_for_status()
        return response

    return await _make_request(url, **httpx_kwargs)


def expand_target_pages(target_pages: str) -> Iterator[int]:
    """Yield all values in target_pages."""
    for target in target_pages.strip().split(","):
        if "-" in target:
            try:
                start, end = map(int, target.strip().split("-"))
                if start > end:
                    raise ValueError
                yield from range(start, end + 1)
            except ValueError as e:
                raise ValueError(f"Invalid page range: {target}") from e
        else:
            try:
                yield int(target)
            except ValueError as e:
                raise ValueError(f"Invalid page number: {target}") from e


def partition_pages(
    pages: Iterable[int], size: int, max_pages: Optional[int] = None
) -> Iterator[str]:
    """Yield partitioned target_pages segments."""
    if size < 1:
        raise ValueError(f"Invalid partition segment size: {size}")
    if max_pages is not None and max_pages < 1:
        raise ValueError("Max pages must be > 0")
    it = iter(pages)
    total = 0
    while max_pages is None or total < max_pages:
        segment = tuple(itertools.islice(it, size))
        if segment:
            targets = []
            for _k, g in itertools.groupby(enumerate(segment), lambda x: x[0] - x[1]):
                group = [item[1] for item in g]
                if len(group) > 1:
                    start, end = group[0], group[-1]
                    group_size = end - start + 1
                    if max_pages is not None and total + group_size > max_pages:
                        end -= total + group_size - max_pages
                        group_size = end - start + 1
                    if group_size > 1:
                        targets.append(f"{start}-{end}")
                    else:
                        targets.append(str(start))
                    total += group_size
                else:
                    targets.append(str(group[0]))
                    total += 1
            yield ",".join(targets)
        else:
            return


def extract_tables_from_json_results(
    json_results: List[dict], download_path: str
) -> List[str]:
    tables = []
    for json_result in json_results:
        pages = json_result["pages"]
        for page in pages:
            page = cast(dict, page)
            items = page.get("items", [])
            if items:
                for i, item in enumerate(items):
                    item = cast(dict, item)
                    if item.get("type", "") == "table" and item.get("csv", ""):
                        savepath = os.path.join(
                            download_path,
                            f"table_{datetime.now().strftime('%Y_%d_%m_%H_%M_%S_%f')[:-3]}.csv",
                        )
                        if Path(savepath).exists():
                            savepath = (
                                savepath.replace(".csv", "_")[0] + str(i) + ".csv"
                            )
                        with open(savepath, "w") as f:
                            f.write(item["csv"])
                        tables.append(savepath)
    return tables



import asyncio
import httpx
import time
from typing import Optional, List, Literal, Union, TYPE_CHECKING
from dataclasses import dataclass
from datetime import datetime
from enum import Enum

from llama_cloud.types import (
    ReportEventItemEventData_Progress,
    ReportMetadata,
    EditSuggestion,
    ReportResponse,
    ReportPlan,
    ReportBlock,
    ReportPlanBlock,
    Report,
)

if TYPE_CHECKING:
    from llama_cloud_services.report.base import LlamaReport


class MessageRole(str, Enum):
    USER = "user"
    ASSISTANT = "assistant"


@dataclass
class Message:
    role: MessageRole
    content: str
    timestamp: datetime


@dataclass
class EditAction:
    block_idx: int
    old_content: str
    new_content: Optional[str]
    action: Literal["approved", "rejected"]
    timestamp: datetime


DEFAULT_POLL_INTERVAL = 5
DEFAULT_TIMEOUT = 600


class ReportClient:
    """Client for operations on a specific report."""

    def __init__(self, report_id: str, name: str, parent_client: "LlamaReport"):
        self.report_id = report_id
        self.name = name
        self._client = parent_client
        self._headers = parent_client.headers
        self._run_sync = parent_client._run_sync
        self._build_url = parent_client._build_url
        self.chat_history: List[Message] = []
        self.edit_history: List[EditAction] = []

    @property
    def aclient(self) -> httpx.AsyncClient:
        return self._client.aclient

    def __str__(self) -> str:
        return f"Report(id={self.report_id}, name={self.name})"

    def __repr__(self) -> str:
        return f"Report(id={self.report_id}, name={self.name})"

    def _get_block_content(self, block: Union[ReportBlock, ReportPlanBlock]) -> str:
        if isinstance(block, ReportBlock):
            return block.template
        elif isinstance(block, ReportPlanBlock):
            return block.block.template
        else:
            raise ValueError(f"Invalid block type: {type(block)}")

    def _get_block_idx(self, block: Union[ReportBlock, ReportPlanBlock]) -> int:
        if isinstance(block, ReportBlock):
            return block.idx
        elif isinstance(block, ReportPlanBlock):
            return block.block.idx
        else:
            raise ValueError(f"Invalid block type: {type(block)}")

    async def aget(self, version: Optional[int] = None) -> ReportResponse:
        """Get this report's details asynchronously."""
        extra_params = []
        if version is not None:
            extra_params.append(f"version={version}")

        url = await self._build_url(f"/api/v1/reports/{self.report_id}", extra_params)

        response = await self.aclient.get(url, headers=self._headers)
        response.raise_for_status()
        return ReportResponse(**response.json())

    def get(self, version: Optional[int] = None) -> ReportResponse:
        """Synchronous wrapper for getting this report's details."""
        return self._run_sync(self.aget(version))

    async def aupdate_report(self, updated_report: Report) -> ReportResponse:
        """Update this report's content asynchronously."""
        url = await self._build_url(f"/api/v1/reports/{self.report_id}")
        response = await self.aclient.patch(
            url, headers=self._headers, json={"content": updated_report.dict()}
        )
        response.raise_for_status()
        return ReportResponse(**response.json())

    def update_report(self, updated_report: Report) -> ReportResponse:
        """Synchronous wrapper for updating this report's content."""
        return self._run_sync(self.aupdate_report(updated_report))

    async def aupdate_plan(
        self,
        action: Literal["approve", "reject", "edit"],
        updated_plan: Optional[ReportPlan] = None,
    ) -> ReportResponse:
        """Update this report's plan asynchronously."""
        if action == "edit" and not updated_plan:
            raise ValueError("updated_plan is required when action is 'edit'")

        url = await self._build_url(
            f"/api/v1/reports/{self.report_id}/plan", [f"action={action}"]
        )

        data = None
        if updated_plan is not None:
            plan_dict = updated_plan.dict()
            plan_dict.pop("generated_at", None)
            data = plan_dict

        if updated_plan is None and action == "edit":
            raise ValueError("updated_plan is required when action is 'edit'")

        response = await self.aclient.patch(url, headers=self._headers, json=data)
        response.raise_for_status()
        return ReportResponse(**response.json())

    def update_plan(
        self,
        action: Literal["approve", "reject", "edit"],
        updated_plan: Optional[ReportPlan] = None,
    ) -> ReportResponse:
        """Synchronous wrapper for updating this report's plan."""
        return self._run_sync(self.aupdate_plan(action, updated_plan))

    async def asuggest_edits(
        self,
        user_query: str,
        auto_history: bool = True,
        chat_history: Optional[List[dict]] = None,
    ) -> List[EditSuggestion]:
        """Get AI suggestions for edits to this report asynchronously.

        Args:
            user_query: The user's request/question about what to edit
            auto_history: Whether to automatically add the user's message to the chat history
            chat_history:
                A list of chat messages to include in the chat history.
                The format being a list of dictionaries with "role" and "content" keys.
        """
        # Add user message to history
        self.chat_history.append(
            Message(role=MessageRole.USER, content=user_query, timestamp=datetime.now())
        )

        # Format chat history with edit summaries
        chat_history_dicts = []
        for msg in self.chat_history[:-1]:  # Exclude current message
            content = msg.content
            if msg.role == MessageRole.USER:
                # Add edit summary for user messages
                edit_summary = self._get_edit_summary_after_message(msg.timestamp)
                if edit_summary:
                    content = f"{content}\n\nActions taken:\n{edit_summary}"

            chat_history_dicts.append({"role": msg.role.value, "content": content})

        # decide whether to include chat history or not
        if chat_history:
            chat_history_dicts = chat_history
        elif auto_history:
            chat_history_dicts = chat_history_dicts
        else:
            chat_history_dicts = []

        # Make the API call
        url = await self._build_url(f"/api/v1/reports/{self.report_id}/suggest_edits")
        data = {"user_query": user_query, "chat_history": chat_history_dicts}

        response = await self.aclient.post(url, headers=self._headers, json=data)
        response.raise_for_status()
        suggestions = response.json()
        suggestions = [EditSuggestion(**suggestion) for suggestion in suggestions]

        # Add assistant response to history
        if suggestions:
            for suggestion in suggestions:
                self.chat_history.append(
                    Message(
                        role=MessageRole.ASSISTANT,
                        content=suggestion.justification,
                        timestamp=datetime.now(),
                    )
                )

        return suggestions

    def suggest_edits(
        self,
        user_query: str,
        auto_history: bool = True,
        chat_history: Optional[List[dict]] = None,
    ) -> List[EditSuggestion]:
        """Synchronous wrapper for getting edit suggestions."""
        return self._run_sync(
            self.asuggest_edits(user_query, auto_history, chat_history)
        )

    async def await_completion(
        self, timeout: int = DEFAULT_TIMEOUT, poll_interval: int = DEFAULT_POLL_INTERVAL
    ) -> Report:
        """Wait for this report to complete processing."""
        start_time = time.time()
        while True:
            report_response = await self.aget()
            status = report_response.status

            if status == "completed":
                return report_response.report
            elif status == "error":
                events = await self.aget_events()
                raise ValueError(f"Report entered error state: {events[-1].msg}")
            elif time.time() - start_time > timeout:
                raise TimeoutError(f"Report did not complete within {timeout} seconds")

            await asyncio.sleep(poll_interval)

    def wait_for_completion(
        self, timeout: int = DEFAULT_TIMEOUT, poll_interval: int = DEFAULT_POLL_INTERVAL
    ) -> Report:
        """Synchronous wrapper for awaiting report completion."""
        return self._run_sync(self.await_completion(timeout, poll_interval))

    async def await_for_plan(
        self, timeout: int = DEFAULT_TIMEOUT, poll_interval: int = DEFAULT_POLL_INTERVAL
    ) -> ReportPlan:
        """Wait for this report's plan to be ready for review."""
        start_time = time.time()
        while True:
            report_metadata = await self.aget_metadata()
            state = report_metadata.state

            if state == "waiting_approval":
                report_response = await self.aget()
                return report_response.plan
            elif state == "error":
                events = await self.aget_events()
                raise ValueError(f"Report entered error state: {events[-1].msg}")
            elif time.time() - start_time > timeout:
                raise TimeoutError(f"Plan was not ready within {timeout} seconds")

            await asyncio.sleep(poll_interval)

    def wait_for_plan(
        self, timeout: int = DEFAULT_TIMEOUT, poll_interval: int = DEFAULT_POLL_INTERVAL
    ) -> ReportPlan:
        """Synchronous wrapper for awaiting plan readiness."""
        return self._run_sync(self.await_for_plan(timeout, poll_interval))

    async def aget_metadata(self) -> ReportMetadata:
        """Get this report's metadata asynchronously."""
        return await self._client.aget_report_metadata(self.report_id)

    def get_metadata(self) -> ReportMetadata:
        """Synchronous wrapper for getting this report's metadata."""
        return self._run_sync(self.aget_metadata())

    async def adelete(self) -> None:
        """Delete this report asynchronously."""
        return await self._client.adelete_report(self.report_id)

    def delete(self) -> None:
        """Synchronous wrapper for deleting this report."""
        return self._run_sync(self.adelete())

    async def aaccept_edit(self, suggestion: EditSuggestion) -> None:
        """Accept a suggested edit.

        Args:
            suggestion: The EditSuggestion to accept, typically from suggest_edits()
        """
        if len(suggestion.blocks) == 0:
            return

        # Determine if we're editing a plan or report based on first block type
        is_plan_edit = isinstance(suggestion.blocks[0], ReportPlanBlock)

        # Get current content
        report_response = await self.aget()
        current_blocks = (
            report_response.plan.blocks
            if is_plan_edit
            else report_response.report.blocks
        )

        # Track the edit
        new_blocks = []
        for edit_block in suggestion.blocks:
            # Find matching block in current content
            old_block = next(
                (
                    b
                    for b in current_blocks
                    if self._get_block_idx(b) == self._get_block_idx(edit_block)
                ),
                None,
            )

            old_content = (
                self._get_block_content(old_block) if old_block else "[No old content]"
            )
            new_content = self._get_block_content(edit_block)

            if is_plan_edit:
                new_queries_str = "\n".join(
                    [
                        f"Field: {q.field}, Prompt: {q.prompt}, Context: {q.context}"
                        for q in edit_block.queries
                    ]
                )
                new_dependency_str = (
                    f"Depends on: {edit_block.dependency}"
                    if edit_block.dependency
                    else ""
                )
                new_content += f"\n\n{new_queries_str}\n{new_dependency_str}"

                if old_block:
                    old_queries_str = "\n".join(
                        [
                            f"Field: {q.field}, Prompt: {q.prompt}, Context: {q.context}"
                            for q in old_block.queries
                        ]
                    )
                    old_dependency_str = (
                        f"Depends on: {old_block.dependency}"
                        if old_block.dependency
                        else ""
                    )
                    old_content += f"\n\n{old_queries_str}\n{old_dependency_str}"

            self.edit_history.append(
                EditAction(
                    block_idx=self._get_block_idx(edit_block),
                    old_content=old_content,
                    new_content=new_content,
                    action="approved",
                    timestamp=datetime.now(),
                )
            )

            # Create updated block
            if is_plan_edit:
                new_blocks.append(
                    ReportPlanBlock(
                        block=ReportBlock(
                            idx=edit_block.block.idx,
                            template=self._get_block_content(edit_block),
                            sources=edit_block.block.sources,
                        ),
                        queries=edit_block.queries,
                        dependency=edit_block.dependency,
                    )
                )
            else:
                new_blocks.append(
                    ReportBlock(
                        idx=edit_block.idx,
                        template=self._get_block_content(edit_block),
                        sources=edit_block.sources,
                    )
                )

        if new_blocks:
            if is_plan_edit:
                # Update plan in place
                plan = report_response.plan

                # Replace edited blocks and add new ones
                for new_block in new_blocks:
                    block_idx = self._get_block_idx(new_block)
                    existing_block_idx = next(
                        (
                            i
                            for i, b in enumerate(plan.blocks)
                            if b.block.idx == block_idx
                        ),
                        None,
                    )

                    if existing_block_idx is not None:
                        # Replace existing block
                        plan.blocks[existing_block_idx] = new_block
                    else:
                        # Add new block to end
                        plan.blocks.append(new_block)

                await self.aupdate_plan("edit", plan)
            else:
                # Update report in place
                report = report_response.report

                # Replace edited blocks and add new ones
                for new_block in new_blocks:
                    block_idx = self._get_block_idx(new_block)
                    existing_block_idx = next(
                        (i for i, b in enumerate(report.blocks) if b.idx == block_idx),
                        None,
                    )

                    if existing_block_idx is not None:
                        # Replace existing block
                        report.blocks[existing_block_idx] = new_block
                    else:
                        # Add new block to end
                        report.blocks.append(new_block)

                await self.aupdate_report(report)

    def accept_edit(self, suggestion: EditSuggestion) -> None:
        """Synchronous wrapper for accepting an edit."""
        return self._run_sync(self.aaccept_edit(suggestion))

    async def areject_edit(self, suggestion: EditSuggestion) -> None:
        """Reject a suggested edit.

        Args:
            suggestion: The EditSuggestion to reject, typically from suggest_edits()
        """
        # Track the rejections
        for edit_block in suggestion.blocks:
            self.edit_history.append(
                EditAction(
                    block_idx=self._get_block_idx(edit_block),
                    old_content=self._get_block_content(edit_block),
                    new_content=None,
                    action="rejected",
                    timestamp=datetime.now(),
                )
            )

    def reject_edit(self, suggestion: EditSuggestion) -> None:
        """Synchronous wrapper for rejecting an edit."""
        return self._run_sync(self.areject_edit(suggestion))

    def _get_edit_summary_after_message(
        self, message_timestamp: datetime
    ) -> Optional[str]:
        """Get a summary of edits that occurred after a specific message."""
        relevant_edits = [
            edit for edit in self.edit_history if edit.timestamp > message_timestamp
        ]

        if not relevant_edits:
            return None

        approved = [edit for edit in relevant_edits if edit.action == "approved"]
        rejected = [edit for edit in relevant_edits if edit.action == "rejected"]

        summary = []

        if approved:
            summary.append("Approved edits:")
            for edit in approved:
                summary.append(
                    f'Block {edit.block_idx}: "{edit.old_content}" -> "{edit.new_content}"'
                )

        if rejected:
            if approved:  # Add spacing if we had approved edits
                summary.append("")
            summary.append("Rejected edits:")
            for edit in rejected:
                summary.append(f'Block {edit.block_idx}: "{edit.old_content}"')

        return "\n".join(summary)

    async def aget_events(
        self, last_sequence: Optional[int] = None
    ) -> List[ReportEventItemEventData_Progress]:
        """Get all events for this report asynchronously.

        Args:
            last_sequence: If provided, only get events after this sequence number

        Returns:
            List of ReportEvent objects
        """
        extra_params = []
        if last_sequence is not None:
            extra_params.append(f"last_sequence={last_sequence}")

        url = await self._build_url(
            f"/api/v1/reports/{self.report_id}/events", extra_params
        )

        response = await self.aclient.get(url, headers=self._headers)
        response.raise_for_status()
        progress_events = []
        for event in response.json():
            if event["event_type"] == "progress":
                progress_events.append(
                    ReportEventItemEventData_Progress(**event["event_data"])
                )

        return progress_events

    def get_events(
        self, last_sequence: Optional[int] = None
    ) -> List[ReportEventItemEventData_Progress]:
        """Synchronous wrapper for getting report events."""
        return self._run_sync(self.aget_events(last_sequence))



import difflib
from pydantic import BaseModel
from typing import Any, Dict, List, Tuple, Type


def check_extra_params(
    model_cls: Type[BaseModel], data: Dict[str, Any]
) -> Tuple[List[str], List[str]]:
    # check if one of the parameters is unused, and warn the user
    model_attributes = set(model_cls.model_fields.keys())
    extra_params = [param for param in data.keys() if param not in model_attributes]

    suggestions: List[str] = []
    if extra_params:
        # for each unused parameter, check if it is similar to a valid parameter and suggest a typo correction, else suggest to check the documentation / update the package
        for param in extra_params:
            similar_params = difflib.get_close_matches(
                param, model_attributes, n=1, cutoff=0.8
            )
            if similar_params:
                suggestions.append(
                    f"'{param}' is not a valid parameter. Did you mean '{similar_params[0]}' instead of '{param}'?"
                )
            else:
                suggestions.append(
                    f"'{param}' is not a valid parameter. Please check the documentation or update the package."
                )

    return extra_params, suggestions



import asyncio
import httpx
import os
import io
from concurrent.futures import ThreadPoolExecutor
from typing import Optional, List, Union, Any, Coroutine, TypeVar
from urllib.parse import urljoin

from llama_cloud.types import ReportMetadata
from llama_cloud_services.report.report import ReportClient

T = TypeVar("T")


class LlamaReport:
    """Client for managing reports and general report operations."""

    def __init__(
        self,
        api_key: Optional[str] = None,
        project_id: Optional[str] = None,
        organization_id: Optional[str] = None,
        base_url: Optional[str] = None,
        timeout: Optional[int] = None,
        async_httpx_client: Optional[httpx.AsyncClient] = None,
    ):
        self.api_key = api_key or os.getenv("LLAMA_CLOUD_API_KEY", None)
        if not self.api_key:
            raise ValueError("No API key provided.")

        self.base_url = base_url or os.getenv(
            "LLAMA_CLOUD_BASE_URL", "https://api.cloud.llamaindex.ai"
        )
        self.timeout = timeout or 60

        # Initialize HTTP clients
        self._aclient = async_httpx_client or httpx.AsyncClient(timeout=self.timeout)

        # Set auth headers
        self.headers = {
            "Authorization": f"Bearer {self.api_key}",
        }

        self.organization_id = organization_id
        self.project_id = project_id
        self._client_params = {
            "timeout": self._aclient.timeout,
            "headers": self._aclient.headers,
            "base_url": self._aclient.base_url,
            "auth": self._aclient.auth,
            "event_hooks": self._aclient.event_hooks,
            "cookies": self._aclient.cookies,
            "max_redirects": self._aclient.max_redirects,
            "params": self._aclient.params,
            "trust_env": self._aclient.trust_env,
        }
        self._thread_pool = ThreadPoolExecutor(
            max_workers=min(10, (os.cpu_count() or 1) + 4)
        )

    @property
    def aclient(self) -> httpx.AsyncClient:
        if self._aclient is None:
            self._aclient = httpx.AsyncClient(**self._client_params)
        return self._aclient

    def _run_sync(self, coro: Coroutine[Any, Any, T]) -> T:
        """Run coroutine in a separate thread to avoid event loop issues"""

        # force a new client for this thread/event loop
        original_client = self._aclient
        self._aclient = None

        def run_coro() -> T:
            async def wrapped_coro() -> T:
                return await coro

            return asyncio.run(wrapped_coro())

        result = self._thread_pool.submit(run_coro).result()

        # restore the original client
        self._aclient = original_client

        return result

    async def _get_default_project(self) -> str:
        response = await self.aclient.get(
            urljoin(str(self.base_url), "/api/v1/projects"), headers=self.headers
        )
        response.raise_for_status()
        projects = response.json()
        default_project = [p for p in projects if p.get("is_default")]
        return default_project[0]["id"]

    async def _build_url(
        self, endpoint: str, extra_params: Optional[List[str]] = None
    ) -> str:
        """Helper method to build URLs with common query parameters."""
        url = urljoin(str(self.base_url), endpoint)

        if not self.project_id:
            self.project_id = await self._get_default_project()

        query_params = []
        if self.organization_id:
            query_params.append(f"organization_id={self.organization_id}")
        if self.project_id:
            query_params.append(f"project_id={self.project_id}")
        if extra_params:
            query_params.extend([p for p in extra_params if p is not None])

        if query_params:
            url += "?" + "&".join(query_params)

        return url

    async def acreate_report(
        self,
        name: str,
        template_instructions: Optional[str] = None,
        template_text: Optional[str] = None,
        template_file: Optional[Union[str, tuple[str, bytes]]] = None,
        input_files: Optional[List[Union[str, tuple[str, bytes]]]] = None,
        existing_retriever_id: Optional[str] = None,
    ) -> ReportClient:
        """Create a new report asynchronously."""
        url = await self._build_url("/api/v1/reports/")
        open_files: List[io.BufferedReader] = []

        data = {"name": name}
        if template_instructions:
            data["template_instructions"] = template_instructions
        if template_text:
            data["template_text"] = template_text
        if existing_retriever_id:
            data["existing_retriever_id"] = str(existing_retriever_id)

        files: List[tuple[str, io.BufferedReader | bytes]] = []
        if template_file:
            if isinstance(template_file, str):
                open_files.append(open(template_file, "rb"))
                files.append(("template_file", open_files[-1]))
            else:
                files.append(("template_file", template_file[1]))

        if input_files:
            for f in input_files:
                if isinstance(f, str):
                    open_files.append(open(f, "rb"))
                    files.append(("files", open_files[-1]))
                else:
                    files.append(("files", f[1]))

        response = await self.aclient.post(
            url, headers=self.headers, data=data, files=files
        )
        try:
            response.raise_for_status()
            report_id = response.json()["id"]
            return ReportClient(report_id, name, self)
        except httpx.HTTPStatusError as e:
            raise ValueError(
                f"Failed to create report: {e.response.text}\nError Code: {e.response.status_code}"
            )
        finally:
            for open_file in open_files:
                open_file.close()

    def create_report(
        self,
        name: str,
        template_instructions: Optional[str] = None,
        template_text: Optional[str] = None,
        template_file: Optional[Union[str, tuple[str, bytes]]] = None,
        input_files: Optional[List[Union[str, tuple[str, bytes]]]] = None,
        existing_retriever_id: Optional[str] = None,
    ) -> ReportClient:
        """Create a new report."""
        return self._run_sync(
            self.acreate_report(
                name=name,
                template_instructions=template_instructions,
                template_text=template_text,
                template_file=template_file,
                input_files=input_files,
                existing_retriever_id=existing_retriever_id,
            )
        )

    async def alist_reports(
        self, state: Optional[str] = None, limit: int = 100, offset: int = 0
    ) -> List[ReportClient]:
        """List all reports asynchronously."""
        params = []
        if state:
            params.append(f"state={state}")
        if limit:
            params.append(f"limit={limit}")
        if offset:
            params.append(f"offset={offset}")

        url = await self._build_url(
            "/api/v1/reports/list",
            extra_params=params,
        )

        response = await self.aclient.get(url, headers=self.headers)
        response.raise_for_status()
        data = response.json()

        return [
            ReportClient(r["report_id"], r["name"], self)
            for r in data["report_responses"]
        ]

    def list_reports(
        self, state: Optional[str] = None, limit: int = 100, offset: int = 0
    ) -> List[ReportClient]:
        """Synchronous wrapper for listing reports."""
        return self._run_sync(self.alist_reports(state, limit, offset))

    async def aget_report(self, report_id: str) -> ReportClient:
        """Get a Report instance for working with a specific report."""
        url = await self._build_url(f"/api/v1/reports/{report_id}")

        response = await self.aclient.get(url, headers=self.headers)
        response.raise_for_status()
        data = response.json()

        return ReportClient(data["report_id"], data["name"], self)

    def get_report(self, report_id: str) -> ReportClient:
        """Synchronous wrapper for getting a report."""
        return self._run_sync(self.aget_report(report_id))

    async def aget_report_metadata(self, report_id: str) -> ReportMetadata:
        """Get metadata for a specific report asynchronously.

        Returns:
            dict containing:
            - id: Report ID
            - name: Report name
            - state: Current report state
            - report_metadata: Additional metadata
            - template_file: Name of template file if used
            - template_instructions: Template instructions if provided
            - input_files: List of input file names
        """
        url = await self._build_url(f"/api/v1/reports/{report_id}/metadata")

        response = await self.aclient.get(url, headers=self.headers)
        response.raise_for_status()
        return ReportMetadata(**response.json())

    def get_report_metadata(self, report_id: str) -> ReportMetadata:
        """Synchronous wrapper for getting report metadata."""
        return self._run_sync(self.aget_report_metadata(report_id))

    async def adelete_report(self, report_id: str) -> None:
        """Delete a specific report asynchronously."""
        url = await self._build_url(f"/api/v1/reports/{report_id}")

        response = await self.aclient.delete(url, headers=self.headers)
        response.raise_for_status()

    def delete_report(self, report_id: str) -> None:
        """Synchronous wrapper for deleting a report."""
        return self._run_sync(self.adelete_report(report_id))



from llama_cloud_services.parse import LlamaParse
from llama_cloud_services.report import ReportClient, LlamaReport
from llama_cloud_services.extract import LlamaExtract, ExtractionAgent
from llama_cloud_services.constants import EU_BASE_URL
from llama_cloud_services.index import (
    LlamaCloudCompositeRetriever,
    LlamaCloudIndex,
    LlamaCloudRetriever,
)

__all__ = [
    "LlamaParse",
    "ReportClient",
    "LlamaReport",
    "LlamaExtract",
    "ExtractionAgent",
    "EU_BASE_URL",
    "LlamaCloudIndex",
    "LlamaCloudRetriever",
    "LlamaCloudCompositeRetriever",
]



from .base import LlamaCloudIndex
from .retriever import LlamaCloudRetriever
from .composite_retriever import (
    LlamaCloudCompositeRetriever,
)

__all__ = [
    "LlamaCloudIndex",
    "LlamaCloudRetriever",
    "LlamaCloudCompositeRetriever",
]



from typing import Any, List, Optional

import httpx
from llama_cloud import (
    CompositeRetrievalMode,
    CompositeRetrievedTextNodeWithScore,
    RetrieverCreate,
    Retriever,
    RetrieverPipeline,
    PresetRetrievalParams,
    ReRankConfig,
)
from llama_cloud.resources.pipelines.client import OMIT

from llama_index.core.base.base_retriever import BaseRetriever
from llama_index.core.constants import DEFAULT_PROJECT_NAME
from llama_index.core.ingestion.api_utils import get_aclient, get_client
from llama_index.core.schema import NodeWithScore, QueryBundle, TextNode
from .base import LlamaCloudIndex
from .api_utils import (
    resolve_project,
    resolve_retriever,
    page_screenshot_nodes_to_node_with_score,
)


class LlamaCloudCompositeRetriever(BaseRetriever):
    def __init__(
        self,
        # retriever identifier
        name: Optional[str] = None,
        retriever_id: Optional[str] = None,
        # project identifier
        project_name: Optional[str] = DEFAULT_PROJECT_NAME,
        project_id: Optional[str] = None,
        organization_id: Optional[str] = None,
        # creation options
        create_if_not_exists: bool = False,
        # connection params
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        app_url: Optional[str] = None,
        timeout: int = 60,
        httpx_client: Optional[httpx.Client] = None,
        async_httpx_client: Optional[httpx.AsyncClient] = None,
        # composite retrieval params
        mode: Optional[CompositeRetrievalMode] = None,
        rerank_top_n: Optional[int] = None,
        rerank_config: Optional[ReRankConfig] = None,
        persisted: Optional[bool] = True,
        **kwargs: Any,
    ) -> None:
        """Initialize the Composite Retriever."""
        # initialize clients
        self._client = get_client(api_key, base_url, app_url, timeout, httpx_client)
        self._aclient = get_aclient(
            api_key, base_url, app_url, timeout, async_httpx_client
        )

        self.project = resolve_project(
            self._client, project_name, project_id, organization_id
        )

        self.name = name
        self.project_name = self.project.name
        self._persisted = persisted

        self.retriever = resolve_retriever(
            self._client, self.project, name, retriever_id, persisted  # type: ignore [arg-type]
        )

        if self.retriever is None and persisted:
            if create_if_not_exists:
                self.retriever = self._client.retrievers.upsert_retriever(
                    project_id=self.project.id,
                    request=RetrieverCreate(name=self.name, pipelines=[]),
                )
            else:
                raise ValueError(
                    f"Retriever with name '{self.name}' does not exist in project."
                )

        # composite retrieval params
        self._mode = mode if mode is not None else OMIT
        self._rerank_top_n = rerank_top_n if rerank_top_n is not None else OMIT
        self._rerank_config = rerank_config if rerank_config is not None else OMIT

        super().__init__(
            callback_manager=kwargs.get("callback_manager"),
            verbose=kwargs.get("verbose", False),
        )

    @property
    def retriever_pipelines(self) -> List[RetrieverPipeline]:
        return self.retriever.pipelines or []  # type: ignore [union-attr]

    def update_retriever_pipelines(
        self, pipelines: List[RetrieverPipeline]
    ) -> Retriever:
        if self._persisted:
            self.retriever = self._client.retrievers.update_retriever(
                self.retriever.id, pipelines=pipelines  # type: ignore [union-attr]
            )
        else:
            # Update in-memory retriever for non-persisted case using copy
            self.retriever = self.retriever.copy(update={"pipelines": pipelines})  # type: ignore [union-attr]
        return self.retriever

    def add_index(
        self,
        index: LlamaCloudIndex,
        name: Optional[str] = None,
        description: Optional[str] = None,
        preset_retrieval_parameters: Optional[PresetRetrievalParams] = None,
    ) -> Retriever:
        name = name or index.name
        preset_retrieval_parameters = (
            preset_retrieval_parameters or index.pipeline.preset_retrieval_parameters
        )
        retriever_pipeline = RetrieverPipeline(
            pipeline_id=index.id,
            name=name,
            description=description,
            preset_retrieval_parameters=preset_retrieval_parameters,
        )
        current_retriever_pipelines_by_name = {
            pipeline.name: pipeline for pipeline in (self.retriever_pipelines or [])
        }
        current_retriever_pipelines_by_name[
            retriever_pipeline.name
        ] = retriever_pipeline
        return self.update_retriever_pipelines(
            list(current_retriever_pipelines_by_name.values())
        )

    def remove_index(self, name: str) -> bool:
        current_retriever_pipeline_names = self.retriever.pipelines or []  # type: ignore [union-attr]
        new_retriever_pipelines = [
            pipeline
            for pipeline in current_retriever_pipeline_names
            if pipeline.name != name
        ]
        if len(new_retriever_pipelines) == len(current_retriever_pipeline_names):
            return False
        self.update_retriever_pipelines(new_retriever_pipelines)
        return True

    async def aupdate_retriever_pipelines(
        self, pipelines: List[RetrieverPipeline]
    ) -> Retriever:
        if self._persisted:
            self.retriever = await self._aclient.retrievers.update_retriever(
                self.retriever.id, pipelines=pipelines  # type: ignore [union-attr]
            )
        else:
            # Update in-memory retriever for non-persisted case using copy
            self.retriever = self.retriever.copy(update={"pipelines": pipelines})  # type: ignore [union-attr]
        return self.retriever

    async def async_add_index(
        self,
        index: LlamaCloudIndex,
        name: Optional[str] = None,
        description: Optional[str] = None,
        preset_retrieval_parameters: Optional[PresetRetrievalParams] = None,
    ) -> Retriever:
        name = name or index.name
        preset_retrieval_parameters = (
            preset_retrieval_parameters or index.pipeline.preset_retrieval_parameters
        )
        retriever_pipeline = RetrieverPipeline(
            pipeline_id=index.id,
            name=name,
            description=description,
            preset_retrieval_parameters=preset_retrieval_parameters,
        )
        current_retriever_pipelines_by_name = {
            pipeline.name: pipeline for pipeline in (self.retriever_pipelines or [])
        }
        current_retriever_pipelines_by_name[
            retriever_pipeline.name
        ] = retriever_pipeline
        return await self.aupdate_retriever_pipelines(
            list(current_retriever_pipelines_by_name.values())
        )

    async def aremove_index(self, name: str) -> bool:
        current_retriever_pipeline_names = self.retriever.pipelines or []  # type: ignore [union-attr]
        new_retriever_pipelines = [
            pipeline
            for pipeline in current_retriever_pipeline_names
            if pipeline.name != name
        ]
        if len(new_retriever_pipelines) == len(current_retriever_pipeline_names):
            return False
        await self.aupdate_retriever_pipelines(new_retriever_pipelines)
        return True

    def _result_nodes_to_node_with_score(
        self, composite_retrieval_node: CompositeRetrievedTextNodeWithScore
    ) -> NodeWithScore:
        return NodeWithScore(
            node=TextNode(
                id=composite_retrieval_node.node.id,
                text=composite_retrieval_node.node.text,
                metadata=composite_retrieval_node.node.metadata,
            ),
            score=composite_retrieval_node.score,
        )

    def _retrieve(
        self,
        query_bundle: QueryBundle,
        mode: Optional[CompositeRetrievalMode] = None,
        rerank_top_n: Optional[int] = None,
        rerank_config: Optional[ReRankConfig] = None,
    ) -> List[NodeWithScore]:
        mode = mode if mode is not None else self._mode

        rerank_top_n = rerank_top_n if rerank_top_n is not None else self._rerank_top_n
        rerank_config = (
            rerank_config if rerank_config is not None else self._rerank_config
        )

        if self._persisted:
            result = self._client.retrievers.retrieve(
                self.retriever.id,  # type: ignore [union-attr]
                mode=mode,
                rerank_top_n=rerank_top_n,
                rerank_config=rerank_config,
                query=query_bundle.query_str,
            )
        else:
            result = self._client.retrievers.direct_retrieve(
                project_id=self.project.id,
                mode=mode,
                rerank_top_n=rerank_top_n,
                rerank_config=rerank_config,
                query=query_bundle.query_str,
                pipelines=self.retriever.pipelines,  # type: ignore [union-attr]
            )
        node_w_scores = [
            self._result_nodes_to_node_with_score(node) for node in result.nodes  # type: ignore [union-attr]
        ]
        image_nodes_w_scores = page_screenshot_nodes_to_node_with_score(
            self._client, result.image_nodes, self.retriever.project_id  # type: ignore [union-attr]
        )
        return sorted(
            node_w_scores + image_nodes_w_scores, key=lambda x: x.score, reverse=True
        )

    async def _aretrieve(
        self,
        query_bundle: QueryBundle,
        mode: Optional[CompositeRetrievalMode] = None,
        rerank_top_n: Optional[int] = None,
        rerank_config: Optional[ReRankConfig] = None,
    ) -> List[NodeWithScore]:
        mode = mode if mode is not None else self._mode

        rerank_top_n = rerank_top_n if rerank_top_n is not None else self._rerank_top_n
        rerank_config = (
            rerank_config if rerank_config is not None else self._rerank_config
        )

        if self._persisted:
            result = await self._aclient.retrievers.retrieve(
                self.retriever.id,  # type: ignore [union-attr]
                mode=mode,
                rerank_config=rerank_config,
                rerank_top_n=rerank_top_n,
                query=query_bundle.query_str,
            )
        else:
            result = await self._aclient.retrievers.direct_retrieve(
                project_id=self.project.id,
                mode=mode,
                rerank_top_n=rerank_top_n,
                rerank_config=rerank_config,
                query=query_bundle.query_str,
                pipelines=self.retriever.pipelines,  # type: ignore [union-attr]
            )
        node_w_scores = [
            self._result_nodes_to_node_with_score(node) for node in result.nodes  # type: ignore [union-attr]
        ]
        image_nodes_w_scores = page_screenshot_nodes_to_node_with_score(
            self._aclient, result.image_nodes, self.retriever.project_id  # type: ignore [union-attr]
        )
        return sorted(
            node_w_scores + image_nodes_w_scores, key=lambda x: x.score, reverse=True
        )



import os
from typing import Any, Dict, Generic, List, Optional, Type

from llama_cloud.client import AsyncLlamaCloud
from tenacity import (
    WrappedFn,
    retry,
    stop_after_attempt,
    wait_exponential,
    retry_if_exception_type,
)
import httpx

from .schema import (
    AgentDataT,
    ComparisonOperator,
    TypedAgentData,
    TypedAgentDataItems,
    TypedAggregateGroup,
    TypedAggregateGroupItems,
)


def agent_data_retry(func: WrappedFn) -> WrappedFn:
    """
    Decorator that adds automatic retry logic to agent data API calls.

    Applies exponential backoff retry strategy for common network-related exceptions:
    - Up to 3 retry attempts
    - Exponential wait time between 0.5s and 10s
    - Retries on timeout, connection, and HTTP status errors

    This ensures resilient API communication in distributed environments where
    temporary network issues or service unavailability may occur.
    """
    return retry(
        stop=stop_after_attempt(3),
        wait=wait_exponential(min=0.5, max=10),
        retry=retry_if_exception_type(
            (httpx.TimeoutException, httpx.ConnectError, httpx.HTTPStatusError)
        ),
    )(func)


def get_default_agent_id() -> str:
    """
    Retrieve the default agent ID from environment variables.

    Returns:
        The value of LLAMA_DEPLOY_DEPLOYMENT_NAME environment variable,
        or None if not set

    Note:
        This provides a convenient way to configure agent ID globally
        via environment variables instead of passing it explicitly
        to each client instance.
    """
    return os.getenv("LLAMA_DEPLOY_DEPLOYMENT_NAME") or "_public"


class AsyncAgentDataClient(Generic[AgentDataT]):
    """
    Async client for managing agent-generated structured data with type safety.

    This client provides a high-level interface for CRUD operations, searching, and
    aggregation of structured data created by agents. It enforces type safety by
    validating all data against a specified Pydantic model type.

    The client is generic over AgentDataT, which must be a Pydantic BaseModel that
    defines the structure of your agent's data output.

    Example:
        ```python
        from pydantic import BaseModel
        from llama_cloud.client import AsyncLlamaCloud
        from llama_cloud_services.beta.agent_data import AsyncAgentDataClient

        class ExtractedPerson(BaseModel):
            name: str
            age: int
            email: str

        # Initialize client
        llama_client = AsyncLlamaCloud(token="your-api-key")
        agent_client = AsyncAgentDataClient(
            client=llama_client,
            type=ExtractedPerson,
            collection="extracted_people",
            agent_url_id="person-extraction-agent"
        )

        # Create data
        person = ExtractedPerson(name="John Doe", age=30, email="john@example.com")
        result = await agent_client.create_agent_data(person)

        # Search data
        results = await agent_client.search(
            filter={"age": {"gt": 25}},
            order_by="data.name",
            page_size=20
        )
        ```

    Type Parameters:
        AgentDataT: Pydantic BaseModel type that defines the structure of agent data
    """

    def __init__(
        self,
        type: Type[AgentDataT],
        collection: str = "default",
        agent_url_id: Optional[str] = None,
        client: Optional[AsyncLlamaCloud] = None,
        token: Optional[str] = None,
        base_url: Optional[str] = None,
    ):
        """
        Initialize the AsyncAgentDataClient.

        Args:
            type: Pydantic BaseModel class that defines the data structure.
                All agent data will be validated against this type.
            collection: Named collection within the agent for organizing data.
                Defaults to "default". Collections allow logical separation of
                different data types or workflows within the same agent.
            agent_url_id: Unique identifier for the agent. This normally appears in the
                url of an agent within the llama cloud platform. If not provided,
                will attempt to use the LLAMA_DEPLOY_DEPLOYMENT_NAME environment
                variable. Data can only be added to an already existing agent in the
                platform.
            client: AsyncLlamaCloud client instance for API communication. If not provided, will
                construct one from the provided api token and base url
            token: Llama Cloud API token. Reads from LLAMA_CLOUD_API_KEY if not provided
            base_url: Llama Cloud API token. Reads from LLAMA_CLOUD_BASE_URL if not provided, and
                defaults to https://api.cloud.llamaindex.ai

        Raises:
            ValueError: If agent_url_id is not provided and the
                LLAMA_DEPLOY_DEPLOYMENT_NAME environment variable is not set

        Note:
            The client automatically applies retry logic to all API calls with
            exponential backoff for timeout, connection, and HTTP status errors.
        """

        self.agent_url_id = agent_url_id or get_default_agent_id()

        self.collection = collection
        if not client:
            client = AsyncLlamaCloud(
                token=token or os.getenv("LLAMA_CLOUD_API_KEY"),
                base_url=base_url or os.getenv("LLAMA_CLOUD_BASE_URL"),
            )
        self.client = client
        self.type = type

    @agent_data_retry
    async def get_item(self, item_id: str) -> TypedAgentData[AgentDataT]:
        raw_data = await self.client.beta.get_agent_data(
            item_id=item_id,
        )
        return TypedAgentData.from_raw(raw_data, validator=self.type)

    @agent_data_retry
    async def create_item(self, data: AgentDataT) -> TypedAgentData[AgentDataT]:
        raw_data = await self.client.beta.create_agent_data(
            agent_slug=self.agent_url_id,
            collection=self.collection,
            data=data.model_dump(),
        )
        return TypedAgentData.from_raw(raw_data, validator=self.type)

    @agent_data_retry
    async def update_item(
        self, item_id: str, data: AgentDataT
    ) -> TypedAgentData[AgentDataT]:
        raw_data = await self.client.beta.update_agent_data(
            item_id=item_id,
            data=data.model_dump(),
        )
        return TypedAgentData.from_raw(raw_data, validator=self.type)

    @agent_data_retry
    async def delete_item(self, item_id: str) -> None:
        await self.client.beta.delete_agent_data(item_id=item_id)

    @agent_data_retry
    async def search(
        self,
        filter: Optional[Dict[str, Dict[ComparisonOperator, Any]]] = None,
        order_by: Optional[str] = None,
        offset: Optional[int] = None,
        page_size: Optional[int] = None,
        include_total: bool = False,
    ) -> TypedAgentDataItems[AgentDataT]:
        """
        Search agent data with filtering, sorting, and pagination.
        Args:
            filter: Filter conditions to apply to the search. Dict mapping field names to FilterOperation objects. Filters only by data fields
                Examples:
                - {"age": {"gt": 18}} - age greater than 18
                - {"status": {"eq": "active"}} - status equals "active"
                - {"tags": {"includes": ["python", "ml"]}} - tags include "python" or "ml"
                - {"created_at": {"gte": "2024-01-01"}} - created after date
                - {"score": {"lt": 100, "gte": 50}} - score between 50 and 100
            order_by: Comma delimited list of fields to sort results by. Can order by standard agent fields like created_at, or by data fields. Data fields must be prefixed with "data.". If ordering desceding, use a " desc" suffix.
                Examples:
                - "data.name desc, created_at" - sort by name in descending order, and then by creation date
            page_size: Maximum number of items to return per page. Defaults to 10.
            offset: Number of items to skip from the beginning. Defaults to 0.
            include_total: Whether to include the total count in the response. Defaults to False to improve performance. It's recommended to only request on the first page.
        """
        raw = await self.client.beta.search_agent_data_api_v_1_beta_agent_data_search_post(
            agent_slug=self.agent_url_id,
            collection=self.collection,
            filter=filter,
            order_by=order_by,
            offset=offset,
            page_size=page_size,
            include_total=include_total,
        )
        return TypedAgentDataItems(
            items=[
                TypedAgentData.from_raw(item, validator=self.type) for item in raw.items
            ],
            has_more=raw.next_page_token is not None,
            total=raw.total_size,
        )

    @agent_data_retry
    async def aggregate(
        self,
        filter: Optional[Dict[str, Dict[ComparisonOperator, Any]]] = None,
        group_by: Optional[List[str]] = None,
        count: Optional[bool] = None,
        first: Optional[bool] = None,
        order_by: Optional[str] = None,
        offset: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> TypedAggregateGroupItems[AgentDataT]:
        """
        Aggregate agent data into groups according to the group_by fields.
        Args:
            filter: Filter conditions to apply to the search. Dict mapping field names to FilterOperation objects. Filters only by data fields
                See search for more details on filtering.
            group_by: List of fields to group by. Groups strictly by equality. Can only group by data fields.
                Examples:
                - ["name"] - group by name
                - ["name", "age"] - group by name and age
            count: Whether to include the count of items in each group.
            first: Whether to include the first item in each group.
            order_by: Comma delimited list of fields to sort results by. See search for more details on ordering.
            offset: Number of groups to skip from the beginning. Defaults to 0.
            page_size: Maximum number of groups to return per page.
        """
        raw = await self.client.beta.aggregate_agent_data_api_v_1_beta_agent_data_aggregate_post(
            agent_slug=self.agent_url_id,
            collection=self.collection,
            page_size=page_size,
            filter=filter,
            order_by=order_by,
            group_by=group_by,
            count=count,
            first=first,
            offset=offset,
        )
        return TypedAggregateGroupItems(
            items=[
                TypedAggregateGroup.from_raw(item, validator=self.type)
                for item in raw.items
            ],
            has_more=raw.next_page_token is not None,
            total=raw.total_size,
        )
